//ENGINE

// GEngine.cpp: implementation of the GEngine class.
//
//////////////////////////////////////////////////////////////////////

#include "ph.h"
#include "GEngine.h"
#include "H3DModel2.h"
#include "GObject.h"
#include "GView.h"
#include "GBackGround.h"
#include "GActiveObject.h"
#include "HCollisions.h"
#include "GRun.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

GEngine::GEngine(GRun *run)
{
	int i;

	for (i = 0; i < MAX_OBJECTS; i++) {
		m_object[i] = NULL;
		m_object_del[i] = NULL;
	}
	m_objects = 0;
	m_objects_del = 0; // pocetr objektu na smazani

	m_run = run; // prirazeni GRun

	// tvorenmi mista pro modely
	for (i = 0; i < MAX_3D_MODELS; i++)
		m_3Dmodel[i] = NULL;
	
	m_paused = false;
	
	m_total_milisec = 0;
	m_total_frames  = 0;
	
	m_collisions = NULL; // nutno nastavit az podle velikosti mapy!!!!
}

/////////////////////////////////////////////////////////////////////////////////////////

GEngine::~GEngine()
{
	int i;

   // ruseni objektu
	while (m_objects > 0)
		m_object[0]->SelfDelete();

	// ruseni modelu
	for (i = 0; i < MAX_3D_MODELS; i++)
		hmodel.RemoveModel(m_3Dmodel[i]);

	if (m_collisions != NULL) 
		delete m_collisions;
	
	// vypiseme prumerny cas jednoho framu
	fprintf(stdout, gettext("Average : %f ms/frame\n"), (float)m_total_milisec/(float)m_total_frames );
}

/////////////////////////////////////////////////////////////////////////////////////////

void GEngine::Render()
{
	m_view.Render();
/*	
	if (m_collisions != NULL) 
		m_collisions->RenderCenters();
*/	
	m_total_frames++;
}

/////////////////////////////////////////////////////////////////////////////////////////

void GEngine::Move()
{

	// otestujem, zda nenastaly podminky pro konec hry
	TestGameEnd();


/*	
/// Ovladani enginu hry
	Uint8 *keystate = SDL_GetKeyState(NULL);

	// posouva kamrou
	if (keystate[SDLK_KP_PLUS]) m_view.m_camera_Z -= 1;
	if (keystate[SDLK_KP_MINUS]) m_view.m_camera_Z += 1;
*/	
}

///////////////////////////////////////////////////////////////////////////////

void GEngine::TimePassed(int milisec)
{
	// zda nahodou neni pauza
	if (m_paused) return;

	// vic, nez 1 sekundu nepovolime - aby se hra "neprehltila"!!!!
	if (milisec > 1000) milisec = 1000;
	
	m_total_milisec += milisec;
	
	m_TimeManager.DoEvents(milisec); // pohne objekty, ktere jsou na rade

	Move(); // aktualizuje stav hry

	DeleteObjects(); // smazeme objekty, ktere si o to rekly


// BERNY: to je tady, aby se zjistovalo, jak rychle se volame - kolik ms trva 1 smycka
// OPTIONALY:	Vypisovani, jak rychle bezi cas
//	fprintf(stderr,"%dms\n",milisec);
}

/////////////////////////////////////////////////////////////////////////////////////////
/*
void GEngine::StartGame()
{

}
*/
/////////////////////////////////////////////////////////////////////////////////////////

bool GEngine::AddObject(GObject *object)
{
	// test na platna data
	if (m_objects >= MAX_OBJECTS || object == NULL) return false;

	// priradime objektu ID v enginu
	object->m_engineID = m_objects;
	
	// pridame do pole objektu
	m_object[m_objects++] = object;

	// pridame do time managera
   m_TimeManager.AddTimeEvent(object);

	// zkusime pridat do kolizniho systemu
	if (m_collisions && IS_AOBJECT(object))
		m_collisions->AddActiveObject((GActiveObject*)object);

   object->m_bValid     = TRUE;
	
	return true;
}

/////////////////////////////////////////////////////////////////////////////////////////

void GEngine::RemoveObject(GObject *object)
{
	// odebrani obejktu z casoveho rizeni udalosti
	m_TimeManager.RemoveTimeEvent(object);
	
	// odebrani z kolizniho sveta, ale jen pokud jde o active objekt
	if (m_collisions != NULL && IS_AOBJECT(object))
		m_collisions->RemoveActiveObject((GActiveObject*) object);

 
	// odebrani z enginu - funguje i pro "posledni" objekt
	m_objects--; // snizime pocet objektu
	// misto objektu na smazani dame posledni objekt
	m_object[object->m_engineID] = m_object[m_objects];
	// upravime engineID
	m_object[object->m_engineID]->m_engineID = object->m_engineID;
	// a posledni pointer pro jistotu NULLnem
	m_object[m_objects] = NULL;
}

/////////////////////////////////////////////////////////////////////////////////////////

void GEngine::DeleteObjects()
{
	// vsechny objekty, ktere si daly pozadavek projdeme a dovolime jim odmazani
	for (int i = 0; i < m_objects_del; i++) 
		if (m_object_del[i] != NULL) {
			m_object_del[i]->SelfDelete();
			m_object_del[i] = NULL;
		}

	m_objects_del = 0; // nastavime pocet objektu na smazani na 0
}

/////////////////////////////////////////////////////////////////////////////////////////

void GEngine::WantBeDeleted(GObject *object_to_delete)
{
	m_object_del[m_objects_del++] = object_to_delete; // pouhe zaregistrovani pozadavku o smazani
}

/////////////////////////////////////////////////////////////////////////////////////////

void GEngine::KeyDown(Uint8 key)
{
	switch (key) {
	case SDLK_p :      // zapauzujem hru
		m_paused = !m_paused;
		break;
	case SDLK_ESCAPE:
		m_run->StartMenu();
		break;
	}
}

/////////////////////////////////////////////////////////////////////////////////////////
/*
inline float abs(float f)
{
	if (f < 0) return -f;
	return f;
};

*/
void GEngine::HitTest(GActiveObject *ob1)
{
	// testovani kolize prihrajeme do kolizniho systemu
	if (m_collisions != NULL)
		m_collisions->HitTest(ob1);
	
}

// zkousi, zda je konec hry (dosli zivoty....)
void GEngine::TestGameEnd()
{

}
