//ENGINE

// GRun.cpp: implementation of the GRun class.
//
//////////////////////////////////////////////////////////////////////

#include "ph.h"
#include "GRun.h"

#include "MMenu.h"
#include "GEngine.h"
#include "GEngineKulic.h"
#include "mvictory.h"
#include "H3DModelBlendQueue.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

GRun::GRun()
{
	m_game_speed = 1;
	m_actual = NULL;
	m_running = true;
	m_loading = false;
}

GRun::~GRun()
{
	// odmazani neodmazaneho
	SAFE_DELETE(m_actual);
	
}

void GRun::Render()
{
	// smazani stare sceny
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	// kresleni aktualniho
	if (m_actual != NULL) m_actual->Render();

	// vyprazdneni bledqueue
	hblendqueue.RenderQueue();
	
	// sem se pripadne muze jeste neco pridat (treba oznaceni betaverze...)
	hfont.printf(20, 460, "Hippo Games 2003 - final beta 0.005");

	// prohodi buffery - zobrazi to, co se vyrenderovalo
	SDL_GL_SwapBuffers();
}

void GRun::TimePassed(int milisec)
{
	// zda se nahodou nenacitalo (menilo m_actual) v tom pripade vynechame jednu smycku
	if (m_loading) {
   	m_loading = false;
    	return;
	}
	// upraveni casu podle rychlosti hry a predani m_actual
	if (m_actual != NULL) {
		if (milisec > 500) milisec = 500; // aby se hra nekousla
		milisec = (int) (m_game_speed * (float)milisec);
		m_actual->TimePassed(milisec);
	}
	
	// a jeste trocha statistiky..
#ifdef WANT_STATISTIC
	Tps_new_frame();
	Tps_timapassed(milisec);
#endif
	
}

// obsluha stisku klavesy
void GRun::KeyDown(Uint8 key)
{
	if (m_actual != NULL) m_actual->KeyDown(key);
}

// obsluha uvolneni klavesy
void GRun::KeyUp(Uint8 key)
{
	if (m_actual != NULL) m_actual->KeyUp(key);
}

// Ukonci program
void GRun::ExitProgram()
{
	// pozzadavek na ukonceni programu - behovou promenou nastavi na false
	m_running = false;
}

// odstratuje hru - jako parametry se predaji kdo co ridi
void GRun::StartGame()
{
	// odstraneni stareho
	SAFE_DELETE(m_actual);
	
	hmodel.DestroyAllModels();
	htexture.DestroyAllTextures();
	hfont.LoadFont("font.bmp");

	// vytvoreni noveho enginu
   GEngineKulic *engine = new GEngineKulic(this);

	// aktivujeme engine - nastavime, aby semu posilaly message
	m_actual = engine;

	// nastavime m_loading
	m_loading = true;	
	
	// predame parametry, jaka hra ma byt (kdo co ovlada)
	if (!engine->StartGame()) StartMenu();
}

// hodi hru do menu
void GRun::StartMenu()
{
	// odstraneni stareho
	SAFE_DELETE(m_actual)
	hmodel.DestroyAllModels();
	htexture.DestroyAllTextures();
	hfont.LoadFont("font.bmp");

	// vytvoreni noveho menu
   m_actual = new MMenu(this);

	// nastavime m_loading
	m_loading = true;
}

// GEngine oznamuje GRunu, ze hra byla dohrana a s jakym vysledkem (alespon jeden byl zabit)
void GRun::GameEnded(st_game_status *game_status)
{
	// odstraneni stareho
	SAFE_DELETE(m_actual)
	htexture.DestroyAllTextures();

	MVictory *vicotry = new MVictory(this, game_status);

	m_actual = vicotry;
	
	// nastavime m_loading
	m_loading = true;
}
