// H3DModel2.h: interface for the H3DObject class.
// prepsany H3Dmodel tak, aby slo animaci uzit

#ifndef _H_3DMODEL2_
#define _H_3DMODEL2_


// struktura pro 1 trojuhelnik. - vybira ze seznamu vrcholu a 
// seznamu souradnic textur
struct tFace
{
	int vertIndex[3];	 // tri vrcholu, ktere tvori trojuhelnik
	int coordIndex[3]; // jejich souradnice v texture
};

// informace o materialu - 
struct tMaterialInfo
{
	char  strName[255];		// jmeno materialu
	char  strFile[255];		// jmeno souboru s texturou
	BYTE  color[3];			// barva (R, G, B)
	int   textureId;			// ID textury v HTextureStore
	bool  bNeedBlenQueue;    // zda je nutne objekt kreslit razenim do H3DModelBlendQueue
} ;

// informace o 1 objektu - cely model se sklada z vice objektum
struct t3DObject 
{
	int  numOfVerts;			// pocet vrcholu v modelu
	int  numOfFaces;			// pocet trojuhelniku
	int  numTexVertex;		// pocet "souradnic v texture"
	int  materialID;			// cislo materialu pro cely objekt
	bool bHasTexture;			// zda ma objekt texturu
	char strName[255];		// jmeno objektu
	bool wantDraw;          // Zda se ma objekt kreslit - nekresli se ty, jejichz jmeno zacina _ - maji specialni vyznam
	CVector3   cubeCenter;  // stred krychle obklopujici objekt
	float      cubeSide;    // delka poloviny hrany krychle - pouziva se na orezavani frustumem
	CVector3   boxMin;      // minimalni souradnice boxu obklopujiciho objekt
	CVector3   boxMax;      // maximalni souradnice boxu

	bool       bHasDisplayList; // zda ma objekt displaylist
	GLuint     displayList;    // pripadne cislo displayListu
	
	CVector3  *pVerts;		// Vrcholy..
	CVector3  *pNormals;		// Normaly
	CVector2  *pTexVerts;	// souradnice v texture
	tFace     *pFaces;		// trojuhelniky
};


// promene na display listy - urcuje, jaky se ma pouzit - dava se funkci LoadModel
#define H3D_DL_NO      0  // bez display listu
#define H3D_DL_OBJECTS 1  // pouze u objektu
#define H3D_DL_MODEL   2  // u celeho modelu najednou


// drzi infomrace o jednom celem modelu
class t3DModel 
{
public:
	t3DModel();
	~t3DModel();

	// zakladni udaje o modelu
	int m_numOfObjects;				 	    // pocet objektu v modelu
	int m_numOfMaterials;					 // pocet materialu v objektu
	vector<tMaterialInfo> m_pMaterials; // vektor materialu
	vector<t3DObject>     m_pObject;	 // vektor objektu
	int                   m_numOfVerts; // celkem trojuhelniku v modelu - pro ucely statistiky
	
	// budeme na frustum testovat i uplne cely model
	CVector3   m_modelCenter;  // stred krychle obklopujici model
	float      m_cubeSide;     // delka poloviny hrany krychle - pouziva se na orezavani frustumem
	
	// zatim se plni jen box
	CVector3   m_boxMin;      // minimalni souradnice boxu obkpolujici model
	CVector3   m_boxMax;      // maximalni souradnice boxu
	CVector3   m_boxCenter;   // stred boxu
	
	
	bool       m_bHasDisplayList; // zda ma cely model displaylist
	GLuint     m_displayList;     // pripadne cislo displayListu
	bool       m_loaded;          // zda je model nacteny
	
	char m_retexturename[255];  // nove jmeno textury
	char m_oldtexturename[255]; // stare jmeno textury - pri nacitani je mozne zamenit 1 texturu na modelu

	// funkce pro praci s modelem
	
	void ZeroModel(); // smazne - nuluje - cely model
	
	// zobrazi objekt - predpoklada se, ze volajici teto funkce nastavil projekcni matici - stred objektu by mel byt v [0,0,0]
	// souradnice jsou dulezite pro orezavani
	void Render(float x, float y, float z);

	// provadi render jednoho objektu - pouziva se u H3DModelBlendQueue i v Render
	// netestuje Frustrum
	void RenderObject(const t3DObject *object);

	// nacte model - rozhodne. zda pouzit 3DS nebo TXT nacitac
   bool LoadModel(const char* filename, int displayLists = H3D_DL_NO);
   bool LoadReTexturedModel(const char* filename, const char* texturename, const char* texturename_old, int displayLists = H3D_DL_NO);

	void GiveObjectToOpenGL(const t3DObject *pObject); // nakresli obekt - preda do OpenGL - pouziva se pri Renderovani i pri stavbe listu
	
	bool Load3DObject3DS(const char *filename); // nacita 3DS model

   void ComputeNormals();      // vypocita normaly k modelu
	void ComputeObjectsCube();  // vypocita krychle k jednotlivym objektum
	void ComputeDisplayLists(int displayLists); // nastavi display listy modelu, aby se dalo pouzit
	
};

// TODO: Mozna by stalo zato, aby se modely ukladaly do display-listu
// teda krome pozadi, ktere by urcite potrebovalo neco jako OCTree (BPSTree ?)

 
// ridi modely a pripadne animaci
class H3DModel  
{
public:

	// zobrazi objekt - predpoklada se, ze volajici teto funkce nastavil projekcni matici - stred objektu by mel byt v [0,0,0]
	// promene x, y a z tam jsou jenom kuli vypoctu frustumu!!!
	void Render(float x, float y, float z);

  // nacte model - rozhodne. zda pouzit 3DS nebo HA3 (animace) nacitac
   bool LoadModel(const char* filename, int displayLists = H3D_DL_NO, bool mirror_x = false, bool mirror_y = false, bool mirror_z = false);
   bool LoadReTexturedModel(const char* filename, const char* texturename, const char* texturename_old, int displayLists = H3D_DL_NO, bool mirror_x = false, bool mirror_y = false, bool mirror_z = false);

	// ....
	H3DModel();
	virtual ~H3DModel();

// 3D model
   t3DModel* m_model;

	bool m_loaded; // zda byl model nacten (je platny)

	// pro ucely animace - animaci lze pouztet i pozadu - zadava se zaporny cas
	void Animate(int milisec); // pohne modelem o X milisekund
	void SetTime(int milisec); // nastavi animacni cas - aby 1 model mohlo pouzivat vic animaci - musi si ale pak sami drzet cas
	void AnimRestart();        // vrati animaci na zacatek
	int    m_loop;               // zda ma animace behat porad dokola
	int    m_animtime;           // celkova doba animace
	int    m_time;               // aktualni cas v animaci - vsechny casove udaje o animaci jsou v ms!
	int    m_models;             // pocet modelu pro animaci
	int    m_frametime;          // cas pro jeden model
	int    m_amodelID;           // ID aktualniho modelu pro render

	// celkove rozmery jeste i tady
	CVector3   m_modelCenter;  // stred krychle obklopujici model
	float      m_cubeSide;     // delka poloviny hrany krychle - pouziva se na orezavani frustumem
	CVector3   m_boxMin;      // minimalni souradnice boxu obkpolujici model
	CVector3   m_boxMax;      // maximalni souradnice boxu
	CVector3   m_boxCenter;   // stred boxu
	

private :

	char m_retexturename[255];  // nove jmeno textury
	char m_oldtexturename[255]; // stare jmeno textury - pri nacitani je mozne zamenit 1 texturu na modelu
	int  m_displayLists;        // zda ma model nejake display listy

	// nacitani konkretnich modelu
	bool Load3DSModel(const char* filename, int displayLists);
	bool LoadHA3Model(const char* filename, int displayLists);

	// prepocita model c podle modelu a, b - p = 0 : c = a; p=1 : c = b;  (p = [0,1])
	void CountAnim(t3DModel* a, t3DModel* b, t3DModel *c, float p);

	// provede mirror modelu podle os
	// provede prepocitani s modelem a pripadne vytvori disply listy
	void MirrorModel(bool mirror_x, bool mirror_y, bool mirror_z);

};


/* struktura HA3 modelu 
	jde vlastne jen o popis animace
	struktura souboru:
------------------
	cas_animace
	pocet_modelu(snimku)_chteny
	pocet_modelu
	jmeno_modelu
   zda se ma cyklit - 0, 1
------------------
tot cele. ve jmenu modelu musi byt %d, kam se budou nacitat modely
modely se cisluji od 0

*/

#endif // _H_3DMODEL2_
