//
// File: H3DModelStore.h
// Created by: berny <berny@seznam.cz>
// Created on: Tue Jun 24 08:06:51 2003
//

#include "ph.h"
#include "H3DModelStore.h"

H3DModelStore hmodel;


H3DModelStore::H3DModelStore()
{
	// nulovani dat
	m_models = 0;
	for (int i = 0; i < MAX_MODELS; i++) m_model[i] = NULL;
}


H3DModelStore::~H3DModelStore()
{
	DestroyAllModels();
}

	
H3DModel* H3DModelStore::LoadModel(const char* filename, int displayLists, bool mirror_x, bool mirror_y, bool mirror_z)
{
	return LoadModelReTextured(filename, "", "", displayLists, mirror_x, mirror_y, mirror_z);
}

H3DModel* H3DModelStore::LoadModelReTextured(const char* filename, const char* texturename, const char* texturename_old, int displayLists, bool mirror_x, bool mirror_y, bool mirror_z)
{

	int i;

	// zmensime pismenka u filename
	char file2[255];
	for (i = 0; filename[i] != '\0'; i++) file2[i] = (filename[i] >= 'A' && filename[i] <= 'Z')? (filename[i] - 'A' + 'a') : filename[i];
	file2[i] = '\0';

	// zmensime pismenka u texturename
	char file3[255];
	for (i = 0; texturename[i] != '\0'; i++) file3[i] = (texturename[i] >= 'A' && texturename[i] <= 'Z')? (texturename[i] - 'A' + 'a') : texturename[i];
	file3[i] = '\0';
	
	// zmensime pismenka u texturename_old
	char file4[255];
	for (i = 0; texturename_old[i] != '\0'; i++) file4[i] = (texturename_old[i] >= 'A' && texturename_old[i] <= 'Z')? (texturename_old[i] - 'A' + 'a') : texturename_old[i];
	file4[i] = '\0';
	
	// hledani jiz jednou vytvoreneho modelu
	for (i = 0; i < m_models; i++)
		if (m_model[i] != NULL && strcmp(file2, m_model[i]->file) == 0 &&
			strcmp(file3, m_model[i]->texture) == 0 && 
			strcmp(file4, m_model[i]->texture_old) == 0 && 
		   m_model[i]->dlist == displayLists &&
			mirror_x == m_model[i]->mirror_x &&  mirror_y == m_model[i]->mirror_y &&  mirror_z == m_model[i]->mirror_z) break;
			
	// model uz nekdo chtel - nebude placat pameti - jen zvysime pocet pouziti
	if (i < m_models) {
		m_model[i]->used++;
#ifdef WANT_STATUS
	printf(gettext("Getting used model %s\n"), filename);
#endif
		return m_model[i]->model;
	}	
	
	// test na volne misto
	if (m_models >= MAX_MODELS) {
		fprintf(stderr, "file: %s, line: %d MAX_MODELS too small - please use larger number.\n", __FILE__, __LINE__);
		return NULL;
	}

#ifdef WANT_STATUS
	printf(gettext("Loading model %s\n"), filename);
#endif
	
	// nacteni noveho 3D modelu
	m_model[m_models] = new st_model;
	m_model[m_models]->model = new H3DModel;
	if (!m_model[m_models]->model->LoadReTexturedModel(file2, file3, file4, displayLists, mirror_x,  mirror_y, mirror_z)) {
		delete m_model[m_models]->model;
		delete m_model[m_models];
		return NULL;
	}
	
	// pridani do seznamu
	m_model[m_models]->used = 1;
	strcpy(m_model[m_models]->file, file2);
	strcpy(m_model[m_models]->texture, file3);
	strcpy(m_model[m_models]->texture_old, file4);
	m_model[m_models]->dlist = displayLists;
	m_model[m_models]->mirror_x = mirror_x;
	m_model[m_models]->mirror_y = mirror_y;
	m_model[m_models]->mirror_z = mirror_z;
	m_models++;
	
	// hotovo
	return m_model[m_models-1]->model;
}
		
 // smazne texturu
void H3DModelStore::RemoveModel(H3DModel* model)
{
	if (model == NULL) return;
		
	// hledani jiz jednou vytvoreneho modelu
	int i;
	for (i = 0; i < m_models; i++)
		if (m_model[i] != NULL && (m_model[i]->model == model)) break;
			
	if (i < m_models) 
		if ( --m_model[i]->used <= 0) DestroyModel(i);
}
  
// Uplne a opravdu zrusi vsechny textury
void H3DModelStore::DestroyAllModels()
{
	int i;
	for (i = 0; i < MAX_MODELS; i++) 
		if (m_model[i] != NULL) {
			delete m_model[i]->model;
			delete m_model[i];
			m_model[i] = NULL;
		}
	m_models = 0;

}
		
// fyzicky znici model
void H3DModelStore::DestroyModel(int ID)
{
	if (m_model[ID] != NULL) {
		delete m_model[ID]->model;
		delete m_model[ID];
		m_models--;
		m_model[ID] = m_model[m_models];
		m_model[m_models] = NULL;
	}
}
