//UTILS

//
// File: HCollisions.h
// Created by: berny <(null)>
// Created on: Wed Mar 19 22:21:26 2003
//

#include "ph.h"

#include "HCollisions.h"
#include "GActiveObject.h"
#include "H3DModel2.h"

 #define __ERROR_CHECKING_ 


HCollisions::HCollisions(float cellszX,  float cellszY, float cellszZ, H3DModel* scenemodel)
{
	// vytvorime vektory, do nich dosadime a posleme to Inicialize();
	CVector3 cellsize;
	cellsize.x = cellszX;
	cellsize.y = cellszY;
	cellsize.z = cellszZ;

	Inicialize(cellsize, scenemodel);
}

HCollisions::HCollisions(float cellszX,  float cellszY, float cellszZ, char* modelname)
{
	// vytvorime vektory, do nich dosadime a posleme to Inicialize();
	CVector3 cellsize;
	cellsize.x = cellszX;
	cellsize.y = cellszY;
	cellsize.z = cellszZ;

//TODO - tady to H3D_DL_OBJECTS by se pak melo zrusit - ted to je pro lepsi nacitani
	// nacteme model
	H3DModel *colissmodel;
	colissmodel = hmodel.LoadModel(modelname, H3D_DL_OBJECTS);
	Inicialize(cellsize, colissmodel);
	hmodel.RemoveModel(colissmodel);
}

HCollisions::HCollisions(CVector3 cellsize, char* modelname)
{
	H3DModel colissmodel;
	colissmodel.LoadModel(modelname);
	Inicialize(cellsize, &colissmodel);
}

HCollisions::HCollisions(CVector3 cellsize, H3DModel* scenemodel)
{
	Inicialize(cellsize, scenemodel);
}

void HCollisions::Inicialize(CVector3 cellsize, H3DModel* scenemodel)
{
	m_cellsize = cellsize; 
	
	
	
	m_minsz[0] = scenemodel->m_boxMin; // prvni pole si ncha puvodni souradnice
	m_minsz[1] = scenemodel->m_boxMin - (cellsize * 0.5); // druhe pole je trochu posunute
	m_maxsz = scenemodel->m_boxMax;
	
	// velikost hraci plochy
	CVector3 size = m_maxsz - m_minsz[0];
	
	// pocet okenek, aby zaplnili hraci plochu
//	m_cells_in_row = 0;
//	while (m_cells_in_row*cellsize < size) m_cells_in_row++;
	m_cells_in_row = (int)(size.x / m_cellsize.x) + 1; // pridame jedno kuli chybam v zaokrouhlovani
	m_cells_in_plane = m_cells_in_row * ((int)(size.z / m_cellsize.z)+1);
	
	// jen pro jistotu
	int i;
	for (i = 0; i < CELL_FIELDS; i++) m_field[i] = NULL;
	
	// velikost celly/2 - polomer kruznice opsane  :-)
	float cell_radius = Magnitude(cellsize) / 3;
	
	// // CELLA
	// vytvorime okenka
	for (i = 0; i < CELL_FIELDS; i++) {
		m_cells[i] = m_cells_in_plane * ((int)(size.y / m_cellsize.y)+1); 
		m_field[i] = new st_colisioncell[m_cells[i]];
		
	// resetovani okenek (aby tam nahodou nebyly nejake blbosti)
		st_colisioncell *cell = m_field[i];
		for (int j = 0; j < m_cells[i]; j++, cell++) {
			cell->objects = 0;
			cell->triangles = 0;

			CVector3 center; // stred celliku
			// vypocitame stred aktualni krychlicky - sileny vypocet :-)
			center.x = m_minsz[i].x + cellsize.x * ((float)(j % m_cells_in_row) + 0.5);
			center.y = m_minsz[i].y + cellsize.y * ((float)((int)(j / m_cells_in_plane)) + 0.5);
			center.z = m_minsz[i].z + cellsize.z * ((float)((int)(j / m_cells_in_row) % m_cells_in_plane) + 0.5);
			

			
			// zkusime do okenka pridat trojuhelniky z modelu
			// tady by se mohlo testovat, zda se vubec objekt s nasim cellikem protina
			for(int i = 0; i < scenemodel->m_model->m_numOfObjects; i++) {
				
				// vezmeme si pointer na aktualne kresleny objekt
				t3DObject *pObject = &scenemodel->m_model->m_pObject[i];
				CVector3 polygon[3]; // souradnice 3 vektoru
			
				// zkusime, zda se vubec s objektem dotyka - NEZBYTNE urychleni testovani :-)
				if (( center.x < pObject->boxMin.x - cell_radius) ||
					 ( center.y < pObject->boxMin.y - cell_radius) ||
					 ( center.z < pObject->boxMin.z - cell_radius) ||
					 ( center.x > pObject->boxMax.x + cell_radius) ||
					 ( center.y > pObject->boxMax.y + cell_radius) ||
					 ( center.z > pObject->boxMax.z + cell_radius)) { 
						 continue;
					 }
				
				// prihlidneme vsechny trojuhelniky objektu a testnem hit
				for (int k = 0; k < pObject->numOfFaces; k++) {
			
					polygon[0] = pObject->pVerts[pObject->pFaces[k].vertIndex[0]];
					polygon[1] = pObject->pVerts[pObject->pFaces[k].vertIndex[1]];
					polygon[2] = pObject->pVerts[pObject->pFaces[k].vertIndex[2]];
					
					// zkusime, zda trojuhelnik koliduje s nasi kouli -> pridame ho do seznamu
					// prestoze mozna pri detekci pomoci koule pridame jeste nejake dalsi, ktere by normalne 
					// byly mimo, nemusi to byt na skodu - v jednom miste, kde se celly dotykaji by se mohlo stat, 
					// ze se kolize nezachyti - diky tomuhle by se stat nemelo 
					if (SpherePolygonCollision(&polygon[0], center, 3, cell_radius)) {
						// pokud je v nasem cellinku, tak ho prodame do seznamu vertexu
						cell->vPolygons.push_back(polygon[0]);
						cell->vPolygons.push_back(polygon[1]);
						cell->vPolygons.push_back(polygon[2]);
						cell->triangles++;
					}
				}
			}
		}
	}
}

HCollisions::~HCollisions()
{
	// smaze vsechna pole
	for (int i = 0; i < CELL_FIELDS; i++)
		if (m_field[i] != NULL)	delete[] m_field[i];
}


void HCollisions::AddActiveObject(GActiveObject *obj)
{
	st_colisioncell *cell;
	
	// prohledneme vsechna kolizni pole a do vsech ho priradime
	for (int i = 0; i < CELL_FIELDS; i++) {
		cell = GetCell(&obj->m_model_position, i);
		obj->m_collissys.cell[i] = NULL;
		AddObjToCell(obj, cell, i);
	}
}

void HCollisions::AddObjToCell(GActiveObject *obj, st_colisioncell *cell, int field)
{
#ifdef __ERROR_CHECKING_
	if (obj == NULL) {
		printf(" %s %d - fatal error \n", __FILE__, __LINE__);
		return;
	}
#endif
	
	// test, zda uz prvek v danem miste vpoli jednou neni
	if (obj->m_collissys.cell[field] != cell && cell != NULL) {
		// pridani do seznamu
		cell->pobject.push_back(obj);
		// aktualizace ID ze seznamu u objektu
		obj->m_collissys.cell_ID[field] = cell->objects;
		obj->m_collissys.cell[field] = cell;
		// zvyseni poctu objektu v cellu
		cell->objects++;
	}
	else 
		if (cell == NULL) {
			obj->m_collissys.cell_ID[field] = 0;
			obj->m_collissys.cell[field] = NULL;
		}
}

// smaze objekt ze ctverecku
void HCollisions::RemoveObjFromCell(GActiveObject *obj, int field)
{

#ifdef __ERROR_CHECKING_
	if (obj == NULL) {
		printf(" %s %d - fatal error \n", __FILE__, __LINE__);
		return;
	}
#endif

	st_colisioncell *cell = obj->m_collissys.cell[field];
	if (cell != NULL) {
		cell->objects--; // snizime pocet objektu

		// poslednimu dame id jako ma ted ruseny
#ifdef __ERROR_CHECKING_
	if (cell->pobject[cell->objects] == NULL) {
		printf(" %s %d - fatal error \n", __FILE__, __LINE__);
		exit(100);
		return;
	}
#endif		
	
		cell->pobject[cell->objects]->m_collissys.cell_ID[field] = obj->m_collissys.cell_ID[field];

		// misto ruseneho dame posledni
		cell->pobject[obj->m_collissys.cell_ID[field]] = cell->pobject[cell->objects];
		// zlikvidujeme posledni policko vektoru
		cell->pobject.pop_back();
	}
}

void HCollisions::RemoveActiveObject(GActiveObject *obj)
{
	// pri odebirai objektu je treba ho vyndat ze vsech poli
	for (int i = 0; i < CELL_FIELDS; i++)
		RemoveObjFromCell(obj, i);
}

// Otestuje objekt po presunuti - vola se z GActiveObject, kdyz se provede pohyb
// tahle fce pak taky pripadne zajisti undomove pri narazu
void HCollisions::HitTest(GActiveObject *obj)
{
	
	// nasatvime undomove na false a zkusime projit vsechna pole
	// funkce TestHit pripadne zmeni na true, pokud se ji neco
	// nebude libit
	m_undomove = false;
	m_changedmovening = false;
	
	int i;

	// projdeme vse a zapamatovavame si celly, aby jsme je pripadne mohli dosadit, 
	// kdyby se objekt pohnul mimo celly

	for (i = 0; i < CELL_FIELDS; i++) {
		// vezmeme ukazatel celliku. Pokud je NULL - objekt by se dostal mimo hraci plochu - 
		// zachovame se jako pri narazu do zdi - undomove + posleme hitted(NUL)
		if ((m_newcell[i] = GetCell(&obj->m_model_position, i)) == NULL) {
			m_undomove = true;
			obj->Hitted(NULL);
		}
		else
			TestHitting(obj, m_newcell[i]);
	}
	
	if (m_undomove) {
		obj->m_position = obj->m_old.m_position; // odsuneme zpet
	}
	else { // posulun se vpred -> otestujeme, zda se nezmenili souradnice cell
		for (i = 0; i < CELL_FIELDS; i++) {
			if (m_newcell[i] != obj->m_collissys.cell[i]) {
				RemoveObjFromCell(obj, i);
				AddObjToCell(obj, m_newcell[i], i);
			}
		}
	}
	
	
}	 	
 /**
	GEngine::HitTest(..);
Projde vsechny objekty (pripadne se objekty utridi a bude se hledat v "okruhu") a pro kazdy
otestuje, zda nenastala kolize.
Neni-li objekt odvozen od GActiveObject, plati (m_type & 1) == 0 a nemuze se ucastnit kolize.
pokud ano, oznamime objektum kolizi
  obj1->Hitted(obj2) a obj1->Hitted(obj2)
funkce Hitted(GObject*) vrati true, pokud bylo teleso "tuhe" a je potreba obstrcit. Poklud
vrati false znamena to, ze kolize se vyresila bez nutnosti menit souradnice (prunikem
objektu nebo znicenim (ne nutne obou) )
if (obj1->Hitted(obj2) && obj1->Hitted(obj2)) {
objekty se od sebe posunou tak, aby nebyly v koliznim prostoru - obj1 se vrati
na puvodni souradnice a podle zakona p = m*v se prohodi vzajemne GActive::m_moveing
}

Je-li hitted volano pro strelu a strela zjisti, ze druhym objektem je lod, ktere chce uskodit,
zavola GShip::LostLives(int lives)
*/	



// do funkce vstupuje ukazatel na testovany objekt a na testovany cell
// nejprve se provede testovani se vsemi objekty v cellu a pote s 
// triangly (s triangly netestuje, pokuz uz byl m_undomove)
// kdyz vznikne kolize s triangly, posle objektu Hitted(NULL)
// pokud byl hit s jinymy objekty, posila mu Hitted(&jiny_objekt)
// a i junemu objektu posle - jiny_objekt.Hitted(obj)
void HCollisions::TestHitting(GActiveObject *obj, st_colisioncell *cell)
{
	// kdyz je objekt want be deleted - nema smysl dal testovat
	//if (obj->m_want_be_deleted) return;
		
//	printf("KOlize : TestHitting\n");
	if (cell == NULL) return;

	// zda ma vubec smysl provadet testovani
	if (cell->objects > 0) {
	
		vector<GActiveObject*>::iterator i; // ukazatel do vektoru
		// a s nim probehneme cely vektor a zkontrolujeme Hitt, ale jen pokud != obj
		
		for (i = cell->pobject.begin(); i != cell->pobject.end(); i++) {
			                           // TODO : tohle je tu jen na testovani!!!
			if (*i != NULL && *i != obj /*&& (*i)->m_active*/) {
				
				// testujeme na naraz - vsechny souradnice musi byt bliz, nez "dovoleno" - taby ten test by asi mel byt jiny
				if (abs(obj->m_model_position.x - (*i)->m_model_position.x) < abs(obj->m_size.x + (*i)->m_size.x) &&
					 abs(obj->m_model_position.y - (*i)->m_model_position.y) < abs(obj->m_size.y + (*i)->m_size.y) &&
					 abs(obj->m_model_position.z - (*i)->m_model_position.z) < abs(obj->m_size.z + (*i)->m_size.z)) {
					// pri narazu zkusime, zda bude treba upravit pozici objektu a jejich smery
					// chceme, aby se obe strany dozvedely o stretu - nelze pouzit if ( && ) kuli optimalizacim kompilatoru - svine jedna :-)
					bool undomove = true; // zda je potreba undomove
					undomove = obj->Hitted((*i)) && undomove; // oznameni o stretu
					undomove = (*i)->Hitted(obj) && undomove;
				
					m_undomove = m_undomove || undomove; // nastavime pozadavek na undomove!!
				
				// SEM SE DOSTANEME< JEN KDYZ JSOU OBA KULICOVE - STRELA SI TO VYRIDI SAMA
					if (undomove && !m_changedmovening && !(*i)->m_want_be_deleted) {
						// odstrceni - prohozeni rychlosti - dokonale pruzna srazka :-)
						CVector3 ob1move = obj->m_moveing;
						// chtelo by to pretizit operatory :-)
				
						// - prohozeni rychlosti v zavislosti na vaze objektu
						obj->m_sliding.x = ((*i)->m_moveing.x*(*i)->m_veight)/obj->m_veight;
						obj->m_sliding.y = ((*i)->m_moveing.y*(*i)->m_veight)/obj->m_veight;
						obj->m_sliding.z = ((*i)->m_moveing.z*(*i)->m_veight)/obj->m_veight;
						(*i)->m_sliding.x = (ob1move.x*obj->m_veight)/(*i)->m_veight;
						(*i)->m_sliding.y = (ob1move.y*obj->m_veight)/(*i)->m_veight;
						(*i)->m_sliding.z = (ob1move.z*obj->m_veight)/(*i)->m_veight;
						
					/*.
						// - prohozeni rychlosti v zavislosti na vaze objektu
						obj->m_moveing.x = ((*i)->m_moveing.x*(*i)->m_veight)/obj->m_veight;
						obj->m_moveing.y = ((*i)->m_moveing.y*(*i)->m_veight)/obj->m_veight;
						obj->m_moveing.z = ((*i)->m_moveing.z*(*i)->m_veight)/obj->m_veight;
						(*i)->m_moveing.x = (ob1move.x*obj->m_veight)/(*i)->m_veight;
						(*i)->m_moveing.y = (ob1move.y*obj->m_veight)/(*i)->m_veight;
						(*i)->m_moveing.z = (ob1move.z*obj->m_veight)/(*i)->m_veight;
						*/
					//	CVector3 ob1move = obj->m_moveing;
						// chtelo by to pretizit operatory :-)
					
						ZeroVector(&(*i)->m_moveing);
						ZeroVector(&obj->m_moveing);
						
						m_changedmovening = true; // aby se nestalo, ze si predaji hybnost tam i zpet
					} 
					
				}
			}
		}
	}
	
	// tady by ted jeste melo prijit testovani na kolizi s polygony!!
	if (!m_undomove) {
		
			// zkusime naraz se vsemi trojuhelniky v cellu
			for(int j = 0; j < cell->triangles*3; j += 3) {
				
				CVector3 polygon[3], pos; // souradnice 3 vektoru
			
				// vezmeme 3 vrcholy trojuhelniku a dame je do jednomo pole zasebe
				polygon[0] = cell->vPolygons[j];
				polygon[1] = cell->vPolygons[j+1];
				polygon[2] = cell->vPolygons[j+2];
				
				pos = obj->m_model_position; // za pozici koule vezmeme pozici panacka
			//	pos.y += 2.0; // a pozuneme ji trochu nahoru - to tady pak neude - je tu jen, aby se dalo chodit i po kolizni mape s "podlahou"
				
				// takhle by to melo asi nejak byt, ale nejsem si jist, zda se plni cubeSide
				//if (SpherePolygonCollision(&polygon[0], obj->m_position, 3, 2 + obj->m_speed /*obj->m_3Dmodel->m_model.cubeSide/2*/)) {
				
				// bereme zatim objekt jako kouli a zkousime ho narazet do trojuhelniku sceny
				// tady by se mohlo udelat treba kolize podle rotovaneho kvadru okolo objektu, ale asi pozdeji...
				if (SpherePolygonCollision(&polygon[0], pos, 3, 1.0f)) {
					m_undomove = true;
					obj->Hitted(NULL);
					break;
				}
			}		
	}
}

// vraci ukazatel na cellA, podle ukazatele na pozici
st_colisioncell* HCollisions::GetCell(CVector3 *position, int field)
{
	int cellID = (int)((int)(position->x - m_minsz[field].x)/m_cellsize.x) + 
	               m_cells_in_row*(int)((position->z - m_minsz[field].z)/m_cellsize.z) + 
	               m_cells_in_plane*(int)((position->y - m_minsz[field].y)/m_cellsize.y);
	
	// pokud je cell mimo rozsah pole, vratime NULL
	if (cellID < 0 || cellID > m_cells[field]-1)	return NULL;
	else return &m_field[field][cellID];
}

// vyrenderuje stredy krychlicek - debug vec
void HCollisions::RenderCenters()
{
	// velikost hraci plochy
	CVector3 size = m_maxsz - m_minsz[0];

	//// CELLA
	// vytvorime okenka
	int i;
	for (i = 0; i < CELL_FIELDS; i++) {
		
	// resetovani okenek (aby tam nahodou nebyly nejake blbosti)
		st_colisioncell *cell = m_field[i];
		for (int j = 0; j < m_cells[i]; j++, cell++) {

			CVector3 center; // stred celliku
			// vypocitame stred aktualni krychlicky - sileny vypocet :-)
			center.x = m_minsz[i].x + m_cellsize.x * ((float)(j % m_cells_in_row) + 0.5);
			center.y = m_minsz[i].y + m_cellsize.y * ((float)((int)(j / m_cells_in_plane)) + 0.5);
			center.z = m_minsz[i].z + m_cellsize.z * ((float)((int)(j / m_cells_in_row) % m_cells_in_plane) + 0.5);

			glBegin(GL_POINTS);
				glVertex3d(center.x, center.y, center.z);
			glEnd();
			
		}
	}	
}


/* Collisions HowTo :

Kolize se v zasade deli na 2 typy. Kolize mezi active ibjekty a kolize mezi
Kolize mezi activeobjekty a "svetem". 

Protoze hra je zatim jen 2D, svet se deli na ctverce, ale casem se bude delit 
na krychle - vyjde to nastejno.
Cela plocha se tedy rozdeli na ctverce primerene velikosti (nejlepsi zjistit
metodou pokusu - ale urcite by mel byt vetsi, nez panacek - nejvetsi z active objekty)
Pak uz se testuje vzdy v ramci jen jednoho maleho ctverecku. Kazdy ctverecek 
tedy obsahuje seznam objektu, ktere v nem jsou, ale take obsahuje seznam trojuhelniku
(polygonu, vertexu) sceny, ktere ho protinaji. Testovani pak bude probihat jen s objekty,
u kterych to ma opravdu smysl. 
Krome testovani kolize bude nejcastejsi operaci scela jiste odebirani a pridavani
objektu (ale nikoliv trojuhelniku - ty se daji jen jednou nazacatku) do ctverecku.
To je zpusobeno tim, jak se objekty hybou. Proto bude i objekt obsahovat pointer primo
na ctverec, ve kterem se nachazi a i index do seznamu obejku v onom ctverci. Usetri
se tim pri odmazavani vyhledavani ctverce a pak linearni hledani v seznamu 
objektu, ktere ctverec v sobe ma.
Nastava problem, poud se objekt dostane blizko k hrane ctverce. Pak se muze stat, ze
by "presahoval" do vedlejsiho ctverce, ale ve vedlejsim ctverci by se o nem nevedelo.
Mohlo by tak dochazet k chybne detekci kolizi. 
1) je mozne hlidat, zda objekt nepresahuje a pak ho zadat do vice ruznych seznamu 
(do vice ctvercu)
2) vytvori se jeste jednopole ctvercu, ktere bude sikmo pousnute oproti puvodnimu. 
Stredy novych ctvercu pak budou lezet presne v miste, kde se setkavaly 4 puvodni
ctverce. Pokud budeme mithodne smulu, muze se stat, ze pokud bude objekt u hranice
dvou ctvercu v obou plochach, kolizi take nezachytime, ale to by se nemelo stavat
prilis - vyzkousime tohle reseni

*/
